<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <title>Dask - Status dashboard</title>
    {{ bokeh_css }}
    {{ bokeh_js }}
    <style>
      @import url("https://fonts.googleapis.com/css?family=Source+Sans+Pro:300,400,700");
      html {
        width: 100%;
        height: 100%;
        background: #FAFAFA;
      }
      body {
        min-height: 100%;
        margin: 0 4%;
        padding: 0px 20px;
        border-left:  1px solid #F0F0F0;
        border-right: 1px solid #F0F0F0;
        background: #FFF;
      }
      @media only screen and (min-width: 90.063em) and (max-width: 120em) {
        /* min-width 1441px and max-width 1920px */
        body {
          margin: 0 20%;
        }
      }
      @media only screen and (min-width: 120.063em) {
        /* min-width 1921px */
        body {
          margin: 0 20%;
        }
      }
      .dashboard {
        clear: both;
      }
      .bk-plot-layout:after {
        clear: both;
        display: block;
        height: 0;
        overflow: hidden;
        visibility: hidden;
        content: ".";
        font-size: 0;
      }
      .navbar img {
        height: 42px;
        float: right;
        margin-bottom: 0px;
        padding: 0px 24px;
      }
      .navbar ul {
          list-style-type: none;
          margin: 0;
          padding: 0;
          overflow: hidden;
          background-color: #EEE;
      }

      .navbar li {
          float: left;
          font-size: 18px;
          transition: .3s background-color;
      }

      .navbar li a {
          display: block;
          color: black;
          text-align: center;
          padding: 11px 16px;
          text-decoration: none;
      }

      .navbar li:hover {
          background-color: #eaaa6d;
      }

      .active {
          background-color: rgba(234, 170, 109, 0.7);
      }

      a {
        color: #1f5396;
        text-decoration: none;
      }
      a:hover,
      a:focus {
        color: #123057;
        text-decoration: underline;
      }
      a:focus {
        outline: 5px auto -webkit-focus-ring-color;
        outline-offset: -2px;
      } .bk-root .bk-toolbar-box .bk-toolbar-right {
        top: 4px;
        right: 4px;
      }
    </style>
  </head>
  <body>
    <div class="navbar">
      <ul>
          <li><a href="http://dask.pydata.org/en/latest/" style="padding:0px 0px;"><img src="https://dask.readthedocs.io/en/latest/_images/dask_horizontal.svg"></a></li>
          {% for page in pages %}
          <li{% if page == active_page %} class="active"{% endif %}><a href="{{ prefix }}/{{ page }}">{{ page|title }}</a></li>
          {% endfor %}
          <li><a href="/info/workers.html">Info</a></li>
      </ul>
    </div>
    <div class="dashboard">
      {{ plot_div }}
    </div>
    {{ plot_script }}
  </body>
</html>
